from __future__ import annotations

import random
import string
from frame_semantic_transformer.data.augmentations.modification_helpers import (
    modify_text_without_changing_length,
)

from frame_semantic_transformer.data.tasks import TaskSample
from .DataAugmentation import DataAugmentation, ProbabilityType


class SimpleMisspellingAugmentation(DataAugmentation):

    max_misspellings_per_sentence: int
    min_misspellings_per_sentence: int

    def __init__(
        self,
        probability: ProbabilityType,
        max_misspellings_per_sentence: int = 10,
        min_misspellings_per_sentence: int = 1,
    ):
        super().__init__(probability)
        self.max_misspellings_per_sentence = max_misspellings_per_sentence
        self.min_misspellings_per_sentence = min_misspellings_per_sentence

    def apply_augmentation(self, task_sample: TaskSample) -> TaskSample:
        def misspell_cb(sentence: str) -> str:
            num_mispellings = random.randint(
                self.min_misspellings_per_sentence, self.max_misspellings_per_sentence
            )
            new_sentence = sentence
            for _ in range(num_mispellings):
                index = random.randint(0, len(sentence) - 1)
                char = sentence[index]
                new_char = char
                if char.isupper():
                    new_char = random.choice(string.ascii_uppercase)
                elif char.islower():
                    new_char = random.choice(string.ascii_lowercase)
                elif char.isdigit():
                    new_char = random.choice(string.digits)
                new_sentence = (
                    new_sentence[:index] + new_char + new_sentence[index + 1 :]
                )
            return new_sentence

        return modify_text_without_changing_length(task_sample, misspell_cb)



from __future__ import annotations
from dataclasses import dataclass
import shutil
from typing import Optional

from transformers import T5ForConditionalGeneration, T5TokenizerFast


class ModelRecorder:

    output_dir: str
    records: list[ModelSaveRecord]

    def __init__(self, output_dir: str):
        self.output_dir = output_dir
        self.records = []

    def save_model(
        self,
        model: T5ForConditionalGeneration,
        tokenizer: T5TokenizerFast,
        epoch: int,
        val_loss: float,
        task_val_metrics: Optional[dict[str, float]] = None,
    ) -> None:
        save_path = self.get_save_path(epoch, val_loss, task_val_metrics)
        model.save_pretrained(save_path)
        tokenizer.save_pretrained(save_path)
        self.records.append(
            ModelSaveRecord(
                epoch=epoch,
                val_loss=val_loss,
                task_val_metrics=task_val_metrics,
                save_path=save_path,
            )
        )

    def remove_non_optimal_models(self) -> None:
        """
        Delete any saved model that doesn't have the best validation loss or
        best validation metric for any task.
        """
        best_val_loss_model = _find_best_val_loss_model(self.records)
        best_val_metric_models = _find_best_val_metric_models(self.records)
        optimal_models = [best_val_loss_model, *best_val_metric_models.values()]
        # clone the list while iterating so we can remove items in the loop
        for record in [*self.records]:
            if record not in optimal_models:
                shutil.rmtree(record.save_path)
                self.records.remove(record)

    def get_save_path(
        self,
        epoch: int,
        val_loss: float,
        task_val_metrics: Optional[dict[str, float]] = None,
    ) -> str:
        filename_parts = [
            f"epoch={epoch}",
            f"val_loss={val_loss}",
        ]
        if task_val_metrics:
            filename_parts.extend(
                [f"{k.replace('-', '_')}={v}" for k, v in task_val_metrics.items()]
            )

        return f"{self.output_dir}/{'-'.join(filename_parts)}"


# moved outside of class for easier testing
def _find_best_val_metric_models(
    records: list[ModelSaveRecord],
) -> dict[str, ModelSaveRecord]:
    best_val_metric_models: dict[str, ModelSaveRecord] = {}
    for record in records:
        if record.task_val_metrics:
            for task, val_metric in record.task_val_metrics.items():
                best_val_model = best_val_metric_models.get(task)
                best_val_metrics = (
                    best_val_model.task_val_metrics if best_val_model else {}
                ) or {}
                best_task_metric = best_val_metrics.get(task)
                if best_task_metric is None or val_metric > best_task_metric:
                    best_val_metric_models[task] = record
    return best_val_metric_models


# moved outside of class for easier testing
def _find_best_val_loss_model(records: list[ModelSaveRecord]) -> ModelSaveRecord:
    best_val_loss_model = records[0]
    for record in records[1:]:
        if record.val_loss < best_val_loss_model.val_loss:
            best_val_loss_model = record
    return best_val_loss_model


@dataclass
class ModelSaveRecord:
    epoch: int
    val_loss: float
    task_val_metrics: Optional[dict[str, float]]
    save_path: str



from __future__ import annotations

from nlpaug.augmenter.word import SynonymAug

from frame_semantic_transformer.data.augmentations.modification_helpers import (
    splice_text,
)
from frame_semantic_transformer.data.augmentations.modification_helpers.splice_text import (
    is_valid_splice,
)

from frame_semantic_transformer.data.tasks import TaskSample
from .DataAugmentation import DataAugmentation, ProbabilityType


class SynonymAugmentation(DataAugmentation):
    """
    Wrapper about nlpaug's SynonymAugmenter
    """

    augmenter: SynonymAug

    def __init__(self, probability: ProbabilityType):
        super().__init__(probability)
        self.augmenter = SynonymAug(aug_max=1, aug_min=1)
        self.augmenter.include_detail = True

    def apply_augmentation(self, task_sample: TaskSample) -> TaskSample:
        def splice_end_punct_cb(
            sentence: str, critical_indices: list[int]
        ) -> tuple[int, int, str] | None:
            _, changes = self.augmenter.augment(sentence)[0]
            if len(changes) == 0:
                return None
            start = changes[0]["orig_start_pos"]
            new_text = changes[0]["new_token"]
            del_len = len(changes[0]["orig_token"])
            if not is_valid_splice(start, del_len, critical_indices):
                return None
            return start, del_len, new_text

        return splice_text(task_sample, splice_end_punct_cb)



from __future__ import annotations
import logging
from typing import Literal, Optional, Union
import pytorch_lightning as pl
import torch
from transformers import T5ForConditionalGeneration, T5TokenizerFast
from pytorch_lightning.loggers import Logger
from pytorch_lightning.callbacks import (
    Callback,
    ModelCheckpoint,
    EarlyStopping,
    TQDMProgressBar,
)
from frame_semantic_transformer.constants import DEFAULT_NUM_WORKERS, MODEL_MAX_LENGTH
from frame_semantic_transformer.data.LoaderDataCache import LoaderDataCache

from frame_semantic_transformer.data.TaskSampleDataset import TaskSampleDataset
from frame_semantic_transformer.data.loaders.framenet17 import (
    Framenet17InferenceLoader,
    Framenet17TrainingLoader,
)
from frame_semantic_transformer.data.loaders.loader import (
    InferenceLoader,
    TrainingLoader,
)
from frame_semantic_transformer.data.tasks_from_annotated_sentences import (
    tasks_from_annotated_sentences,
)
from frame_semantic_transformer.training.TrainingDataModule import TrainDataModule
from frame_semantic_transformer.training.TrainingModelWrapper import (
    TrainingModelWrapper,
)

logger = logging.getLogger(__name__)


def train(
    base_model: str = "t5-base",
    batch_size: int = 8,
    max_epochs: int = 10,
    use_gpu: bool = torch.cuda.is_available(),
    output_dir: str = "outputs",
    early_stopping_patience_epochs: int = 0,  # 0 to disable early stopping feature
    precision: Union[Literal[64, 32, 16], Literal["64", "32", "16", "bf16"]] = 32,
    lr: float = 1e-4,
    lr_gamma: float = 1.0,
    num_workers: int = DEFAULT_NUM_WORKERS,
    save_only_last_epoch: bool = False,
    balance_tasks: bool = True,
    max_task_duplication_factor: int = 2,
    skip_initial_epochs_validation: int = 0,
    inference_loader: Optional[InferenceLoader] = None,
    training_loader: Optional[TrainingLoader] = None,
    pl_callbacks: Optional[list[Callback]] = None,
    pl_loggers: Optional[list[Logger]] = None,
    resume_from_checkpoint: Optional[str] = None,
    remove_non_optimal_models: bool = True,
) -> tuple[T5ForConditionalGeneration, T5TokenizerFast]:
    device = torch.device("cuda" if use_gpu else "cpu")
    logger.info("loading base T5 model")
    model = T5ForConditionalGeneration.from_pretrained(base_model).to(device)
    tokenizer = T5TokenizerFast.from_pretrained(
        base_model, model_max_length=MODEL_MAX_LENGTH
    )
    if not inference_loader:
        inference_loader = Framenet17InferenceLoader()
    loader_cache = LoaderDataCache(inference_loader)
    if not training_loader:
        training_loader = Framenet17TrainingLoader()

    model.config.training_loader = training_loader.name()
    model.config.inference_loader = inference_loader.name()

    inference_loader.setup()
    training_loader.setup()

    logger.info("loading train/test/val datasets")
    training_data = training_loader.load_training_data()
    validation_data = training_loader.load_validation_data()
    test_data = training_loader.load_test_data()

    train_dataset = TaskSampleDataset(
        tasks_from_annotated_sentences(training_data, loader_cache),
        tokenizer,
        balance_tasks=balance_tasks,
        max_task_duplication_factor=max_task_duplication_factor,
        augmentations=training_loader.get_augmentations(),
    )
    val_dataset = TaskSampleDataset(
        tasks_from_annotated_sentences(validation_data, loader_cache),
        tokenizer,
        balance_tasks=False,
    )
    test_dataset = TaskSampleDataset(
        tasks_from_annotated_sentences(test_data, loader_cache),
        tokenizer,
        balance_tasks=False,
    )

    data_module = TrainDataModule(
        train_dataset=train_dataset,
        val_dataset=val_dataset,
        test_dataset=test_dataset,
        batch_size=batch_size,
        num_workers=num_workers,
    )

    model_wrapper = TrainingModelWrapper(
        model,
        tokenizer,
        lr=lr,
        lr_gamma=lr_gamma,
        output_dir=output_dir,
        save_only_last_epoch=save_only_last_epoch,
        skip_initial_epochs_validation=skip_initial_epochs_validation,
        loader_cache=loader_cache,
        remove_non_optimal_models=remove_non_optimal_models,
    )

    # add callbacks
    callbacks: list[Callback] = [TQDMProgressBar(refresh_rate=5)]
    if pl_callbacks:
        callbacks.extend(pl_callbacks)

    if early_stopping_patience_epochs > 0:
        early_stop_callback = EarlyStopping(
            monitor="val_loss",
            min_delta=0.00,
            patience=early_stopping_patience_epochs,
            verbose=True,
            mode="min",
        )
        callbacks.append(early_stop_callback)

    checkpoint_callback = ModelCheckpoint(
        monitor="val_loss",
        dirpath=output_dir,
        filename="best-checkpoint",
        save_top_k=1,
        mode="min",
    )
    callbacks.append(checkpoint_callback)

    # prepare trainer
    trainer = pl.Trainer(
        callbacks=callbacks,
        max_epochs=max_epochs,
        accelerator="gpu" if use_gpu else "cpu",
        devices=1 if use_gpu else "auto",
        precision=precision,
        log_every_n_steps=1,
        logger=pl_loggers or True,
        resume_from_checkpoint=resume_from_checkpoint,
    )

    logger.info("beginning training")

    # fit trainer
    trainer.fit(model_wrapper, data_module)

    return model, tokenizer



from __future__ import annotations
import re

from nltk.stem import PorterStemmer

from .load_propbank_frames import load_propbank_frames

from .ensure_propbank_downloaded import ensure_propbank_downloaded


from frame_semantic_transformer.data.frame_types import Frame
from ..loader import InferenceLoader


base_stemmer = PorterStemmer()

LOW_PRIORITY_LONGER_LUS = {"back", "down", "make", "take", "have", "into", "come"}


class Propbank34InferenceLoader(InferenceLoader):
    """
    Inference loader for Propbank 3.1 data
    """

    def setup(self) -> None:
        ensure_propbank_downloaded()

    def strict_frame_elements(self) -> bool:
        """
        Propbank only lists core roles, not all roles, so we can't enforce strict frame elements
        """
        return False

    def load_frames(self) -> list[Frame]:
        """
        Load the full list of frames to be used during inference
        """
        return load_propbank_frames()

    def normalize_lexical_unit_text(self, lu: str) -> str:
        """
        Normalize a lexical unit like "takes.v" to "take".
        """
        normalized_lu = lu.lower().replace("_", " ")
        normalized_lu = re.sub(r"\.[a-zA-Z]+$", "", normalized_lu)
        normalized_lu = re.sub(r"[^a-z0-9 ]", "", normalized_lu)
        return base_stemmer.stem(normalized_lu.strip())

    def prioritize_lexical_unit(self, lu: str) -> bool:
        """
        Check if the lexical unit is relatively rare, so that it should be considered "high information"
        """
        norm_lu = self.normalize_lexical_unit_text(lu)
        return len(norm_lu) >= 4 and norm_lu not in LOW_PRIORITY_LONGER_LUS



from __future__ import annotations
import re
from frame_semantic_transformer.data.augmentations.modification_helpers import (
    splice_text,
)
from frame_semantic_transformer.data.augmentations.modification_helpers.splice_text import (
    is_valid_splice,
)

from frame_semantic_transformer.data.tasks import TaskSample
from .DataAugmentation import DataAugmentation

REMOVE_END_PUNCT_RE = r"\s*[.?!]\s*$"


class RemoveEndPunctuationAugmentation(DataAugmentation):
    def apply_augmentation(self, task_sample: TaskSample) -> TaskSample:
        def splice_end_punct_cb(
            sentence: str, critical_indices: list[int]
        ) -> tuple[int, int, str] | None:
            match = re.search(REMOVE_END_PUNCT_RE, sentence)
            if match is None:
                return None
            start, end = match.span()
            del_len = end - start
            if not is_valid_splice(start, del_len, critical_indices):
                return None
            return start, del_len, ""

        return splice_text(task_sample, splice_end_punct_cb)



__version__ = "0.8.0"

from .FrameSemanticTransformer import (
    FrameSemanticTransformer,
    DetectFramesResult,
    FrameElementResult,
    FrameResult,
)
from .data.loaders.loader import InferenceLoader, TrainingLoader
from .data.frame_types import (
    FrameAnnotatedSentence,
    FrameAnnotation,
    FrameElementAnnotation,
    Frame,
)

__all__ = (
    "FrameSemanticTransformer",
    "DetectFramesResult",
    "FrameElementResult",
    "FrameResult",
    "InferenceLoader",
    "TrainingLoader",
    "FrameAnnotatedSentence",
    "FrameAnnotation",
    "FrameElementAnnotation",
    "Frame",
)



from __future__ import annotations
import re

from nltk.stem import (
    PorterStemmer,
    LancasterStemmer,
    SnowballStemmer,
    WordNetLemmatizer,
)
from nltk.corpus import framenet as fn

from .ensure_framenet_downloaded import ensure_framenet_downloaded
from .ensure_wordnet_downloaded import ensure_wordnet_downloaded
from frame_semantic_transformer.data.frame_types import Frame
from ..loader import InferenceLoader


porter_stemmer = PorterStemmer()
lancaster_stemmer = LancasterStemmer()
snowball_stemmer = SnowballStemmer("english")
wordnet_lemmatizer = WordNetLemmatizer()

WORDNET_LEMMATIZER_POS = ["a", "r", "n", "v", "s"]


LOW_PRIORITY_LONGER_LUS = {"back", "down", "make", "take", "have", "into", "come"}


class Framenet17InferenceLoader(InferenceLoader):
    """
    Inference loader for FrameNet 1.7 data
    """

    def setup(self) -> None:
        ensure_framenet_downloaded()
        ensure_wordnet_downloaded()

    def load_frames(self) -> list[Frame]:
        """
        Load the full list of frames to be used during inference
        """
        frames = []
        for raw_frame in fn.frames():
            frame = Frame(
                name=raw_frame.name,
                core_elements=[
                    name for (name, fe) in raw_frame.FE.items() if fe.coreType == "Core"
                ],
                non_core_elements=[
                    name for (name, fe) in raw_frame.FE.items() if fe.coreType != "Core"
                ],
                lexical_units=[lu for lu in raw_frame.lexUnit.keys()],
            )
            frames.append(frame)
        return frames

    def normalize_lexical_unit_text(self, lu: str) -> str | set[str]:
        """
        Normalize a lexical unit like "takes.v" to "take".
        """
        normalized_lu = lu.lower()
        normalized_lu = re.sub(r"\.[a-zA-Z]+$", "", normalized_lu)
        normalized_lu = re.sub(r"[^a-z0-9 ]", "", normalized_lu)
        normalized_lu = normalized_lu.strip()
        norm_lus = {
            porter_stemmer.stem(normalized_lu),
            lancaster_stemmer.stem(normalized_lu),
            snowball_stemmer.stem(normalized_lu),
        }
        # try every possible part of speech for the wordnet lemmatizer
        for pos in WORDNET_LEMMATIZER_POS:
            norm_lus.add(wordnet_lemmatizer.lemmatize(normalized_lu, pos=pos))
        return norm_lus

    def prioritize_lexical_unit(self, lu: str) -> bool:
        """
        Check if the lexical unit is relatively rare, so that it should be considered "high information"
        """
        norm_lu = self.normalize_lexical_unit_text(lu)
        return len(norm_lu) >= 4 and norm_lu not in LOW_PRIORITY_LONGER_LUS



from __future__ import annotations

import os

from transformers import T5ForConditionalGeneration, T5TokenizerFast

from frame_semantic_transformer.data.loaders.loader import (
    InferenceLoader,
    TrainingLoader,
)
from frame_semantic_transformer.training.evaluate_model import evaluate_model

from frame_semantic_transformer.training.find_best_val_model_paths import (
    find_best_val_model_paths,
)


def evaluate_best_val_models(
    outputs_dir: str,
    inference_loader: InferenceLoader,
    training_loader: TrainingLoader,
    batch_size: int,
    num_workers: int,
) -> None:
    """
    Helper to run the evaluation on the best models after training
    """

    best_models = find_best_val_model_paths(outputs_dir)
    evaluated_models = set()
    for key, output_name in best_models.items():
        if output_name in evaluated_models:
            continue
        print(f"Best {key} model: {output_name}")
        model_path = os.path.join(outputs_dir, output_name)
        model = T5ForConditionalGeneration.from_pretrained(model_path)
        tokenizer = T5TokenizerFast.from_pretrained(model_path)
        print(f"Evaluating model: {output_name}")
        evaluate_model(
            model,
            tokenizer,
            inference_loader=inference_loader,
            training_loader=training_loader,
            batch_size=batch_size,
            num_workers=num_workers,
        )
        evaluated_models.add(output_name)



from __future__ import annotations
from typing import Callable, Sequence

from frame_semantic_transformer.data.tasks import TaskSample

from .DataAugmentation import DataAugmentation


def chain_augmentations(
    augmentations: Sequence[DataAugmentation],
) -> Callable[[TaskSample], TaskSample]:
    def chained_augmentation(input: TaskSample) -> TaskSample:
        chained_input = input
        for augmentation in augmentations:
            chained_input = augmentation(chained_input)
        return chained_input

    return chained_augmentation



from __future__ import annotations
import logging
import os
import argparse

from frame_semantic_transformer.training.train import train

DEFAULT_NUM_WORKERS = os.cpu_count() or 2
logger = logging.getLogger(__name__)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Train a new frame semantic transformer model using FrameNet 1.7"
    )
    parser.add_argument(
        "--base-model",
        default="t5-base",
        help="The HuggingFace T5 model to use as a starting point, default t5-base",
    )
    parser.add_argument("--learning-rate", default=1e-4, type=float)
    parser.add_argument("--batch-size", type=int, default=8, help="default 8")
    parser.add_argument("--epochs", type=int, default=10, help="default 10")
    parser.add_argument(
        "--output-dir",
        default="outputs",
        help="dir where output models will be saved after each epoch, default ./outputs",
    )
    args = parser.parse_args()

    logger.setLevel(logging.INFO)

    train(
        base_model=args.base_model,
        batch_size=args.batch_size,
        lr=args.learning_rate,
        max_epochs=args.epochs,
        output_dir=args.output_dir,
    )



from __future__ import annotations

import nltk
from glob import glob
from os import path
from xml.etree import ElementTree


from frame_semantic_transformer.data.frame_types import Frame


def load_propbank_frames() -> list[Frame]:
    """
    Load the full list of frames to be used during inference
    """
    dataset_path = nltk.data.find("corpora/propbank-frames-3.4.0").path
    frames_paths = glob(path.join(dataset_path, "frames", "*.xml"))
    frames = []
    for frame_path in frames_paths:
        with open(frame_path, "r") as frame_file:
            etree = ElementTree.parse(frame_file).getroot()
            raw_frames = etree.findall("predicate/roleset")
            for raw_frame in raw_frames:
                frame = Frame(
                    name=raw_frame.attrib["id"],
                    core_elements=[
                        f"ARG{role.attrib['n']}-{role.attrib['f']}"
                        for role in raw_frame.findall("roles/role")
                    ],
                    non_core_elements=[],
                    lexical_units=[
                        f"{alias.text}.{alias.attrib['pos']}"
                        for alias in raw_frame.findall("aliases/alias")
                    ],
                )
                frames.append(frame)
    return frames



from __future__ import annotations
from dataclasses import dataclass
import re
from typing import Sequence
from frame_semantic_transformer.data.LoaderDataCache import LoaderDataCache

from frame_semantic_transformer.data.frame_types import FrameElementAnnotation

from .ArgumentsExtractionTask import ArgumentsExtractionTask, split_output_fe_spans
from .TaskSample import TaskSample


@dataclass
class ArgumentsExtractionSample(TaskSample):
    task: ArgumentsExtractionTask
    frame_elements: list[FrameElementAnnotation]

    # -- input / target / eval for training --

    def get_target(self) -> str:
        return " | ".join(
            [f"{element} = {text}" for element, text in self.labeled_frame_elements]
        )

    @staticmethod
    def evaluate_prediction(
        prediction_outputs: Sequence[str],
        target: str,
        input: str,
        loader_cache: LoaderDataCache,
    ) -> tuple[float, float, float]:
        # based on argid eval in sesame (labeled_eval)
        # from https://github.com/swabhs/open-sesame/blob/master/sesame/evaluation.py

        frame = re.sub(r"\s+\|.*$", "", input.replace("ARGS ", "")).strip()

        true_pos = 0.0
        false_pos = 0.0
        false_neg = 0.0
        target_spans = split_output_fe_spans(target)
        prediction_spans = ArgumentsExtractionTask.parse_output(
            prediction_outputs, loader_cache
        )

        for target_span in target_spans:
            score = get_eval_score(frame, target_span[0], loader_cache)
            if target_span in prediction_spans:
                true_pos += score
            else:
                false_neg += score

        for prediction_span in prediction_spans:
            if prediction_span not in target_spans:
                score = get_eval_score(frame, prediction_span[0], loader_cache)
                false_pos += score

        return (true_pos, false_pos, false_neg)

    # -- helper properties --

    @property
    def labeled_frame_elements(self) -> list[tuple[str, str]]:
        """
        Return a list of tuples of the form (frame_element, text) for all frame elements, instead of indices
        """
        return [
            (fe.name, self.task.text[fe.start_loc : fe.end_loc])
            for fe in self.frame_elements
        ]


def get_eval_score(frame: str, element: str, loader_cache: LoaderDataCache) -> float:
    if element in loader_cache.get_frame(frame).non_core_elements:
        return 0.5
    return 1.0



from __future__ import annotations

from typing import Callable, Union
from abc import ABC, abstractmethod
from random import uniform

from frame_semantic_transformer.data.tasks.TaskSample import TaskSample


ProbabilityType = Union[float, Callable[[TaskSample], float]]


class DataAugmentation(ABC):
    """
    Base class for data augmentations on training data
    """

    probability: ProbabilityType

    def __init__(self, probability: ProbabilityType):
        self.probability = probability

    def __call__(self, task_sample: TaskSample) -> TaskSample:
        """
        randomly apply this augmentation in proportion to self.probability
        """
        rand_val = uniform(0, 1.0)
        if rand_val > self.get_probability(task_sample):
            return task_sample
        return self.apply_augmentation(task_sample)

    def get_probability(self, task_sample: TaskSample) -> float:
        if callable(self.probability):
            return self.probability(task_sample)
        return self.probability

    @abstractmethod
    def apply_augmentation(self, task_sample: TaskSample) -> TaskSample:
        """
        Main logic for subclasses to implement
        """
        pass



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from typing import Dict, Iterable, Optional, Tuple, cast
import torch
from transformers import T5ForConditionalGeneration, T5TokenizerFast

from frame_semantic_transformer.constants import (
    MODEL_MAX_LENGTH,
    MODEL_REVISION,
    OFFICIAL_RELEASES,
)
from frame_semantic_transformer.data.LoaderDataCache import LoaderDataCache
from frame_semantic_transformer.data.data_utils import chunk_list, marked_string_to_locs
from frame_semantic_transformer.data.loaders.loader import InferenceLoader
from frame_semantic_transformer.data.loaders.framenet17 import Framenet17InferenceLoader
from frame_semantic_transformer.predict import batch_predict
from frame_semantic_transformer.data.tasks import (
    ArgumentsExtractionTask,
    FrameClassificationTask,
    Task,
    TriggerIdentificationTask,
)

# has the form {sentence: {trigger_loc: (frame, {frame_element: text})}}
ResultsAccumulator = Dict[str, Dict[int, Tuple[str, Dict[str, str]]]]


@dataclass
class FrameElementResult:
    name: str
    text: str


@dataclass
class FrameResult:
    name: str
    trigger_location: int
    frame_elements: list[FrameElementResult]


@dataclass
class DetectFramesResult:
    sentence: str
    trigger_locations: list[int]
    frames: list[FrameResult]


class FrameSemanticTransformer:
    _model: T5ForConditionalGeneration | None = None
    _tokenizer: T5TokenizerFast | None = None
    model_path: str
    model_revision: str | None = None
    device: torch.device
    batch_size: int
    predictions_per_sample: int
    loader_cache: LoaderDataCache

    def __init__(
        self,
        model_name_or_path: str = "base",
        use_gpu: bool = torch.cuda.is_available(),
        batch_size: int = 8,
        predictions_per_sample: int = 5,
        inference_loader: Optional[InferenceLoader] = None,
    ):
        self.model_path = model_name_or_path
        if model_name_or_path in OFFICIAL_RELEASES:
            self.model_path = f"chanind/frame-semantic-transformer-{model_name_or_path}"
            self.model_revision = MODEL_REVISION
        self.device = torch.device("cuda" if use_gpu else "cpu")
        self.batch_size = batch_size
        self.predictions_per_sample = predictions_per_sample
        loader = inference_loader or Framenet17InferenceLoader()
        self.loader_cache = LoaderDataCache(loader)

    def setup(self) -> None:
        """
        Initialize the model and tokenizer, and download models / files as needed
        If this is not called explicitly it will be lazily called before inference
        """
        self._model = T5ForConditionalGeneration.from_pretrained(
            self.model_path, revision=self.model_revision
        ).to(self.device)
        self._tokenizer = T5TokenizerFast.from_pretrained(
            self.model_path,
            revision=self.model_revision,
            model_max_length=MODEL_MAX_LENGTH,
        )
        self.loader_cache.setup()
        self._validate_loader()

    def _validate_loader(self) -> None:
        """
        Helper to ensure that the loader being used matches the one used to train the model
        otherwise results will be potentially nonsensical
        """
        loader = self.loader_cache.loader
        config = self.model.config
        expected_loader_name = (
            config.inference_loader
            if hasattr(config, "inference_loader")
            else loader.name()
        )
        if expected_loader_name != loader.name():
            raise ValueError(
                f"Model {self.model_path} was trained with inference loader {expected_loader_name} "
                f"but you are using {loader.name()}. Please pass in the correct loader as 'inference_loader' "
                f"when initializing FrameSemanticTransfomer."
            )

    @property
    def model(self) -> T5ForConditionalGeneration:
        if not self._model:
            self.setup()
        return cast(T5ForConditionalGeneration, self._model)

    @property
    def tokenizer(self) -> T5TokenizerFast:
        if not self._tokenizer:
            self.setup()
        return cast(T5TokenizerFast, self._tokenizer)

    def _batch_predict(self, inputs: list[str]) -> list[str]:
        """
        helper to avoid needing to repeatedly pass in the same params every call to predict
        """
        return batch_predict(
            self.model,
            self.tokenizer,
            inputs,
            num_beams=self.predictions_per_sample,
            num_return_sequences=self.predictions_per_sample,
        )

    def detect_frames(self, sentence: str) -> DetectFramesResult:
        return self.detect_frames_bulk([sentence])[0]

    def _collate_results(
        self, results_acc: ResultsAccumulator
    ) -> dict[str, DetectFramesResult]:
        """
        helper to turn the results accumulator into DetectFramesResults
        """
        results: dict[str, DetectFramesResult] = {}
        for sentence, frame_data in results_acc.items():
            frame_results: list[FrameResult] = []
            for trigger_loc, (frame, elements) in frame_data.items():
                frame_results.append(
                    FrameResult(
                        name=frame,
                        trigger_location=trigger_loc,
                        frame_elements=[
                            FrameElementResult(element, text)
                            for element, text in elements.items()
                        ],
                    )
                )
            results[sentence] = DetectFramesResult(
                sentence,
                trigger_locations=list(frame_data.keys()),
                frames=frame_results,
            )
        return results

    def detect_frames_bulk(self, sentences: Iterable[str]) -> list[DetectFramesResult]:
        tasks_queue: list[Task] = []
        # slowly build up results from each task as they complete
        results_acc: ResultsAccumulator = defaultdict(dict)
        # T5 doesn't necessarily have to output the original sentence, even though it's supposed to
        # This map just keeps track of the original sentence for each output sentence so we can match them up
        parsed_sentences_map: dict[str, str] = {}
        for sentence in sentences:
            tasks_queue.append(TriggerIdentificationTask(text=sentence))
        while len(tasks_queue) > 0:
            batch = tasks_queue[: self.batch_size]
            tasks_queue = tasks_queue[self.batch_size :]
            batch_results = self._batch_predict([task.get_input() for task in batch])
            for task, preds in zip(
                batch, chunk_list(batch_results, self.predictions_per_sample)
            ):
                if isinstance(task, TriggerIdentificationTask):
                    # first identify triggers
                    result = task.parse_output(preds, self.loader_cache)
                    parsed_sent, trigger_locs = marked_string_to_locs(result)
                    parsed_sentences_map[task.text] = parsed_sent
                    for trigger_loc in trigger_locs:
                        tasks_queue.append(
                            FrameClassificationTask(
                                text=task.text,
                                trigger_loc=trigger_loc,
                                loader_cache=self.loader_cache,
                            )
                        )
                elif isinstance(task, FrameClassificationTask):
                    # then identify frames
                    frame = task.parse_output(preds, self.loader_cache)
                    if frame:
                        results_acc[task.text][task.trigger_loc] = (frame, {})
                        tasks_queue.append(
                            ArgumentsExtractionTask(
                                text=task.text,
                                trigger_loc=task.trigger_loc,
                                frame=frame,
                                loader_cache=self.loader_cache,
                            )
                        )
                elif isinstance(task, ArgumentsExtractionTask):
                    # finally identify frame elements
                    frame_element_tuples = task.parse_output(preds, self.loader_cache)
                    results_acc[task.text][task.trigger_loc][1].update(
                        frame_element_tuples
                    )
        results_map = self._collate_results(results_acc)
        results = []
        for sentence in sentences:
            mapped_sentence = parsed_sentences_map[sentence]
            if mapped_sentence in results_map:
                results.append(results_map[mapped_sentence])
            else:
                results.append(
                    DetectFramesResult(sentence, trigger_locations=[], frames=[])
                )
        return results



import nltk
from nltk.downloader import Package


# NLTK only has v1.0 of PropBank, so hackily create a NLTK package and download v3.1
propbank34 = Package(
    id="propbank-frames-3.4.0",
    url="https://github.com/propbank/propbank-frames/archive/refs/tags/v3.4.0.zip",
    name="Proposition Bank Corpus 3.4",
    checksum="e563f8c9912d53ed7e709455746875e5",
    subdir="corpora",
    size=9484561,
    unzipped_size=29870379,
)


def ensure_propbank_downloaded() -> None:
    try:
        nltk.data.find("corpora/propbank-frames-3.4.0.zip")
    except LookupError:
        nltk.download(propbank34)



from .Propbank34InferenceLoader import Propbank34InferenceLoader
from .Propbank34TrainingLoader import Propbank34TrainingLoader


__all__ = [
    "Propbank34InferenceLoader",
    "Propbank34TrainingLoader",
]



from __future__ import annotations

from nlpaug.augmenter.char import KeyboardAug

from frame_semantic_transformer.data.augmentations.modification_helpers import (
    modify_text_without_changing_length,
)
from frame_semantic_transformer.data.tasks import TaskSample
from .DataAugmentation import DataAugmentation, ProbabilityType


class KeyboardAugmentation(DataAugmentation):
    """
    Wrapper about nlpaug's KeyboardAugmenter
    Attempts to make spelling mistakes similar to what a user might make
    """

    augmenter: KeyboardAug

    def __init__(self, probability: ProbabilityType):
        super().__init__(probability)
        self.augmenter = KeyboardAug(
            include_special_char=False, aug_char_p=0.1, aug_word_p=0.1
        )
        self.augmenter.include_detail = True

    def apply_augmentation(self, task_sample: TaskSample) -> TaskSample:
        def augment_sent(sentence: str) -> str:
            # this augmentation removes spaces around punctuation, so just manually do the changes
            _, changes = self.augmenter.augment(sentence)[0]
            new_sentence = sentence
            for change in changes:
                # sometimes this augmenter changes token lengths, which we don't want
                # just skip the changes if that happens
                if len(change["orig_token"]) != len(change["new_token"]):
                    return new_sentence
                if change["orig_start_pos"] != change["new_start_pos"]:
                    return new_sentence
                start_pos = change["orig_start_pos"]
                end_pos = change["orig_start_pos"] + len(change["orig_token"])
                new_sentence = (
                    new_sentence[:start_pos]
                    + change["new_token"]
                    + new_sentence[end_pos:]
                )
            return new_sentence

        return modify_text_without_changing_length(task_sample, augment_sent)



from .Framenet17InferenceLoader import Framenet17InferenceLoader
from .Framenet17TrainingLoader import Framenet17TrainingLoader


__all__ = [
    "Framenet17InferenceLoader",
    "Framenet17TrainingLoader",
]



from __future__ import annotations
from dataclasses import dataclass
from typing import Sequence

from transformers import T5TokenizerFast

from frame_semantic_transformer.data.LoaderDataCache import LoaderDataCache
from frame_semantic_transformer.data.data_utils import standardize_punct


from .Task import Task


@dataclass
class ArgumentsExtractionTask(Task):
    text: str
    trigger_loc: int
    frame: str
    loader_cache: LoaderDataCache

    @staticmethod
    def get_task_name() -> str:
        return "args_extraction"

    def get_input(self) -> str:
        frame = self.loader_cache.get_frame(self.frame)
        core_elements = frame.core_elements
        non_core_elements = frame.non_core_elements
        # put core elements in front
        elements = [*core_elements, *non_core_elements]
        return f"ARGS {self.frame} | {' '.join(elements)} : {self.trigger_labeled_text}"

    @staticmethod
    def parse_output(
        prediction_outputs: Sequence[str], loader_cache: LoaderDataCache
    ) -> list[tuple[str, str]]:
        raw_elements = split_output_fe_spans(prediction_outputs[0])
        processed_elements = []
        for element, text in raw_elements:
            fixed_element = loader_cache.standardize_element_name(element)
            if fixed_element is not None:
                processed_elements.append((fixed_element, text))
        return processed_elements

    @property
    def trigger_labeled_text(self) -> str:
        pre_span = self.text[0 : self.trigger_loc]
        post_span = self.text[self.trigger_loc :]
        # TODO: handle these special chars better
        return standardize_punct(f"{pre_span}*{post_span}")


def split_output_fe_spans(output: str) -> list[tuple[str, str]]:
    """
    Split an output like "Agent = He | Destination = to the store" into a list of elements and values, like:
    [("Agent", "He"), ("Destination", "to the store")]
    """
    outputs: list[tuple[str, str]] = []
    for span in T5TokenizerFast.clean_up_tokenization(output).split("|"):
        parts = span.strip().split("=")
        if len(parts) == 1:
            # invalid output - just skip this
            continue
        else:
            outputs.append(
                (
                    parts[0].strip(),
                    parts[1].strip(),
                )
            )
    return outputs



from __future__ import annotations

import os


MODEL_MAX_LENGTH = 512
OFFICIAL_RELEASES = ["base", "small"]  # TODO: small, large
MODEL_REVISION = "v0.2.0"
PADDING_LABEL_ID = -100
DEFAULT_NUM_WORKERS = os.cpu_count() or 2



from __future__ import annotations
from collections import defaultdict
import random
from typing import Any, Callable, Optional, Sequence
import torch
from torch.utils.data import Dataset
from transformers import T5TokenizerFast
from frame_semantic_transformer.constants import MODEL_MAX_LENGTH, PADDING_LABEL_ID

from frame_semantic_transformer.data.augmentations import (
    DataAugmentation,
    chain_augmentations,
)
from frame_semantic_transformer.data.tasks import TaskSample


MAX_TARGET_LEN = 512


class TaskSampleDataset(Dataset[Any]):
    samples: Sequence[TaskSample]
    augmentation: Optional[Callable[[TaskSample], TaskSample]] = None
    tokenizer: T5TokenizerFast

    def __init__(
        self,
        samples: Sequence[TaskSample],
        tokenizer: T5TokenizerFast,
        balance_tasks: bool = False,
        seed: int = 42,
        max_task_duplication_factor: int = 2,
        augmentations: Optional[list[DataAugmentation]] = None,
    ):
        self.samples = samples
        if balance_tasks:
            self.samples = balance_tasks_by_type(
                samples, seed=seed, max_duplication_factor=max_task_duplication_factor
            )
        self.tokenizer = tokenizer
        if augmentations:
            self.augmentation = chain_augmentations(augmentations)

    def __len__(self) -> int:
        return len(self.samples)

    def __getitem__(self, index: int) -> dict[str, Any]:
        sample = self.samples[index]

        input_ids, attention_mask, labels = self.parse_sample(sample)

        return {
            "input_ids": input_ids.squeeze(),
            "attention_mask": attention_mask.squeeze(),
            "labels": labels,
            "task": sample.get_task_name(),
        }

    def parse_sample(
        self, sample: TaskSample
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:

        aug_sample = sample
        if self.augmentation:
            aug_sample = self.augmentation(sample)
        input = aug_sample.get_input()
        target = aug_sample.get_target()

        input_encoding = self.tokenizer(
            input,
            padding="max_length",
            max_length=MODEL_MAX_LENGTH,
            truncation=True,
            return_tensors="pt",
        )
        input_ids, attention_mask = (
            input_encoding.input_ids,
            input_encoding.attention_mask,
        )
        output_encoding = self.tokenizer(
            target,
            padding="max_length",
            max_length=MAX_TARGET_LEN,
            truncation=True,
        )
        labels = torch.tensor(output_encoding.input_ids)
        labels[labels == self.tokenizer.pad_token_id] = PADDING_LABEL_ID

        return (input_ids, attention_mask, labels)


def balance_tasks_by_type(
    samples: Sequence[TaskSample],
    max_duplication_factor: int = 2,
    seed: int = 42,
) -> Sequence[TaskSample]:
    """
    try to force an approximate balance of task types by repeating tasks of uncommon types
    """
    counts_by_type: dict[str, int] = defaultdict(int)
    for sample in samples:
        counts_by_type[sample.get_task_name()] += 1
    max_task_count = max(counts_by_type.values())
    balanced_samples: list[TaskSample] = []
    for sample in samples:
        duplication_factor = int(
            max_task_count / counts_by_type[sample.get_task_name()]
        )
        # duplicate each sample in proportion to how few tasks of this type are in the original mix
        for _ in range(min(max_duplication_factor, duplication_factor)):
            balanced_samples.append(sample)
    random.Random(seed).shuffle(balanced_samples)
    return balanced_samples



from __future__ import annotations
from dataclasses import dataclass
from typing import Sequence

from frame_semantic_transformer.data.LoaderDataCache import LoaderDataCache

from .Task import Task


@dataclass
class TriggerIdentificationTask(Task):
    text: str

    # -- input / target for training --

    @staticmethod
    def get_task_name() -> str:
        return "trigger_identification"

    def get_input(self) -> str:
        return f"TRIGGER: {self.text}"

    @staticmethod
    def parse_output(
        prediction_outputs: Sequence[str], _loader_cache: LoaderDataCache
    ) -> str:
        return prediction_outputs[0]



from __future__ import annotations
import random

from frame_semantic_transformer.data.augmentations.modification_helpers import (
    splice_text,
)
from frame_semantic_transformer.data.tasks import TaskSample
from .DataAugmentation import DataAugmentation
from .modification_helpers.get_sample_text import get_sample_text


LATEX_QUOTES = ["``", "''"]
STANDARD_QUOTE = '"'
ALL_QUOTES = LATEX_QUOTES + [STANDARD_QUOTE]


class DoubleQuotesAugmentation(DataAugmentation):
    def apply_augmentation(self, task_sample: TaskSample) -> TaskSample:
        sample_text = get_sample_text(task_sample)

        # if standard quotes are used, convert to latex quotes, and vice versa
        to_latex = STANDARD_QUOTE in sample_text
        from_quotes = [STANDARD_QUOTE] if to_latex else LATEX_QUOTES
        to_quotes = LATEX_QUOTES if to_latex else [STANDARD_QUOTE]

        updated_sample = task_sample
        while count_instances(sample_text, from_quotes) > 0:
            quote, start_loc = find_first_instance(sample_text, from_quotes)
            try:
                updated_sample = splice_text(
                    updated_sample,
                    lambda _text, _critical_indices: (
                        start_loc,
                        len(quote),
                        random.choice(to_quotes),
                    ),
                )
                sample_text = get_sample_text(updated_sample)
            except ValueError:
                # The splice failed, so just return the sample
                return updated_sample

        return updated_sample


def count_instances(text: str, substrings: list[str]) -> int:
    return sum(text.count(substring) for substring in substrings)


def find_first_instance(text: str, substrings: list[str]) -> tuple[str, int]:
    """
    Find the first instance of any of the substrings in the text. Returns the substring and the
    start location of the substring.
    """
    for substring in substrings:
        start_loc = text.find(substring)
        if start_loc >= 0:
            return substring, start_loc
    raise ValueError(f"Could not find any of {substrings} in {text}")



from .Task import Task
from .TaskSample import TaskSample
from .ArgumentsExtractionSample import ArgumentsExtractionSample
from .ArgumentsExtractionTask import ArgumentsExtractionTask
from .FrameClassificationSample import FrameClassificationSample
from .FrameClassificationTask import FrameClassificationTask
from .TriggerIdentificationSample import TriggerIdentificationSample
from .TriggerIdentificationTask import TriggerIdentificationTask

__all__ = (
    "Task",
    "TaskSample",
    "ArgumentsExtractionSample",
    "ArgumentsExtractionTask",
    "FrameClassificationSample",
    "FrameClassificationTask",
    "TriggerIdentificationSample",
    "TriggerIdentificationTask",
)



from __future__ import annotations
from collections import defaultdict
import json
from os import path, makedirs
from typing import Any

import numpy as np
import pytorch_lightning as pl
import torch
from torch.optim.lr_scheduler import ExponentialLR
from transformers import AdamW, T5ForConditionalGeneration, T5TokenizerFast

from frame_semantic_transformer.data.LoaderDataCache import LoaderDataCache
from frame_semantic_transformer.data.data_utils import trim_batch
from frame_semantic_transformer.training.ModelRecorder import ModelRecorder
from .evaluate_batch import TaskEvalResults, calc_eval_metrics, evaluate_batch


class TrainingModelWrapper(pl.LightningModule):
    """
    Based on https://github.com/Shivanandroy/simpleT5/blob/main/simplet5/simplet5.py
    """

    lr: float
    model: T5ForConditionalGeneration
    tokenizer: T5TokenizerFast
    trainer: pl.Trainer
    output_dir: str
    save_only_last_epoch: bool
    skip_initial_epochs_validation: int
    loader_cache: LoaderDataCache
    val_metrics: dict[str, float] | None
    lr_gamma: float
    log_eval_failures: bool
    model_recorder: ModelRecorder

    def __init__(
        self,
        model: T5ForConditionalGeneration,
        tokenizer: T5TokenizerFast,
        loader_cache: LoaderDataCache,
        lr: float = 1e-4,
        output_dir: str = "outputs",
        save_only_last_epoch: bool = False,
        skip_initial_epochs_validation: int = 0,
        lr_gamma: float = 1.0,
        log_eval_failures: bool = False,
        remove_non_optimal_models: bool = True,
    ):
        super().__init__()
        self.lr = lr
        self.model = model
        self.tokenizer = tokenizer
        self.loader_cache = loader_cache
        self.output_dir = output_dir
        self.save_only_last_epoch = save_only_last_epoch
        self.skip_initial_epochs_validation = skip_initial_epochs_validation
        self.val_metrics = None
        self.lr_gamma = lr_gamma
        self.log_eval_failures = log_eval_failures
        self.model_recorder = ModelRecorder(output_dir)
        self.remove_non_optimal_models = remove_non_optimal_models

    def forward(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        return self.model(*args, **kwargs)

    def _step(self, batch: Any) -> Any:
        with torch.no_grad():
            input_ids, attention_mask, labels = trim_batch(
                batch["input_ids"], batch["attention_mask"], batch["labels"]
            )
        return self(
            input_ids=input_ids,
            attention_mask=attention_mask,
            labels=labels,
        )

    def training_step(self, batch: Any, _batch_idx: int) -> Any:  # type: ignore
        output = self._step(batch)
        loss = output.loss
        self.log(
            "train_loss",
            loss,
            prog_bar=True,
            logger=True,
            on_epoch=True,
            on_step=True,
            batch_size=len(batch["input_ids"]),
        )
        return loss

    def validation_step(self, batch: Any, _batch_idx: int) -> Any:  # type: ignore
        output = self._step(batch)
        loss = output.loss
        if self.current_epoch < self.skip_initial_epochs_validation:
            return {"loss": loss}
        metrics = evaluate_batch(self.model, self.tokenizer, batch, self.loader_cache)
        self.log(
            "val_loss",
            loss,
            prog_bar=True,
            logger=True,
            on_epoch=True,
            on_step=True,
            batch_size=len(batch["input_ids"]),
        )
        return {"loss": loss, "metrics": metrics}

    def test_step(self, batch: Any, _batch_idx: int) -> Any:  # type: ignore
        output = self._step(batch)
        loss = output.loss
        metrics = evaluate_batch(self.model, self.tokenizer, batch, self.loader_cache)
        self.log(
            "test_loss",
            loss,
            prog_bar=True,
            logger=True,
            batch_size=len(batch["input_ids"]),
        )
        return {"loss": loss, "metrics": metrics}

    def configure_optimizers(self) -> tuple[list[AdamW], list[ExponentialLR]]:
        optimizer = AdamW(self.parameters(), lr=self.lr)
        scheduler = ExponentialLR(optimizer, gamma=self.lr_gamma, verbose=True)
        return [optimizer], [scheduler]

    def training_epoch_end(self, training_step_outputs: list[Any]) -> None:
        """save tokenizer and model on epoch end"""
        self.average_training_loss = np.round(
            torch.mean(torch.stack([x["loss"] for x in training_step_outputs])).item(),
            4,
        )
        self.log("train_loss", self.average_training_loss)
        if (
            not self.save_only_last_epoch
            or self.current_epoch == (self.trainer.max_epochs or 0) - 1
        ):
            self.model_recorder.save_model(
                self.model,
                self.tokenizer,
                epoch=self.current_epoch,
                val_loss=self.average_validation_loss,
                task_val_metrics=self.val_metrics,
            )
            if self.remove_non_optimal_models:
                self.model_recorder.remove_non_optimal_models()

    def validation_epoch_end(self, validation_step_outputs: list[Any]) -> None:
        losses = [out["loss"].cpu() for out in validation_step_outputs]
        self.average_validation_loss = np.round(
            torch.mean(torch.stack(losses)).item(),
            4,
        )
        self.log("val_loss", self.average_validation_loss)
        if self.current_epoch < self.skip_initial_epochs_validation:
            # no validation metrics to calculate in this epoch, just return early
            return

        metrics = merge_metrics([out["metrics"] for out in validation_step_outputs])
        self.val_metrics = {}
        for task_name, results in metrics.items():
            scores = calc_eval_metrics(results.scores)
            name = f"val_{task_name}_f1"
            f_score = scores["f_score"]
            self.val_metrics[name] = f_score
            self.log(name, f_score)
            self.log(f"val_{task_name}_recall", scores["recall"])
            self.log(f"val_{task_name}_precision", scores["precision"])

        if self.log_eval_failures:
            log_eval_failures(
                self.output_dir,
                f"val_{self.current_epoch}_eval_failures.json",
                metrics,
            )

    def test_epoch_end(self, test_step_outputs: list[Any]) -> None:
        average_test_loss = np.round(
            torch.mean(torch.stack([x["loss"] for x in test_step_outputs])).item(),
            4,
        )
        self.log("test_loss", average_test_loss)
        metrics = merge_metrics([out["metrics"] for out in test_step_outputs])
        for task_name, results in metrics.items():
            scores = calc_eval_metrics(results.scores)
            self.log(f"test_{task_name}_f1", scores["f_score"])
            self.log(f"test_{task_name}_recall", scores["recall"])
            self.log(f"test_{task_name}_precision", scores["precision"])
        if self.log_eval_failures:
            log_eval_failures(
                self.output_dir,
                f"test_{self.current_epoch}_eval_failures.json",
                metrics,
            )


def merge_metrics(
    metrics: list[dict[str, TaskEvalResults]]
) -> dict[str, TaskEvalResults]:
    merged_metrics: dict[str, TaskEvalResults] = defaultdict(TaskEvalResults)
    for metric in metrics:
        for task_name, eval_results in metric.items():
            merged_metrics[task_name].scores.false_pos += eval_results.scores.false_pos
            merged_metrics[task_name].scores.false_neg += eval_results.scores.false_neg
            merged_metrics[task_name].scores.true_pos += eval_results.scores.true_pos
            merged_metrics[task_name].false_negatives += eval_results.false_negatives
            merged_metrics[task_name].false_positives += eval_results.false_positives
    return merged_metrics


def log_eval_failures(
    output_dir: str, filename: str, eval_results: dict[str, TaskEvalResults]
) -> None:
    serialized_results = {
        task: results.serialize() for task, results in eval_results.items()
    }
    failures_file = path.join(output_dir, filename)
    makedirs(output_dir, exist_ok=True)
    with open(failures_file, "w+") as f:
        json.dump(serialized_results, f, indent=2)



from __future__ import annotations
from dataclasses import replace
from typing import Callable

from frame_semantic_transformer.data.tasks import (
    ArgumentsExtractionSample,
    FrameClassificationSample,
    TaskSample,
    TriggerIdentificationSample,
)


def is_valid_splice(
    start_loc: int,
    delete_num: int,
    critical_indices: list[int],
) -> bool:
    """
    Helper to check if a splice is valid. A splice is valid if it does not delete any of the critical indices.
    """
    for index in critical_indices:
        if index >= start_loc and index < start_loc + delete_num:
            return False
    return True


def splice_text(
    task_sample: TaskSample,
    modify_text_cb: Callable[[str, list[int]], tuple[int, int, str] | None],
) -> TaskSample:
    """
    Helper to modify the text of a TaskSample that may change the length of the text. This is
    a more complex augmentation since it requires rewriting indices, and can potentially break
    the sample. This is loosely modified on the `splice()` function from javascript, where
    a start position is given, followed by the number of chars to remove, and then the new
    string to insert.

    This takes the task sample and a lambda function. The lambda funtion that takes the text of the
    task sample and a list of critical indices to the task, which must not be deleted during the splice.
    The lambda returns a tuple of the start position, the number of chars to remove, and the new string
    to insert.
    """

    def modify_text(
        text: str, critical_indices: list[int]
    ) -> tuple[str, Callable[[int], int]]:
        modify_results = modify_text_cb(text, critical_indices)
        if modify_results is None:
            return text, lambda i: i
        start_loc, delete_num, insert_text = modify_results
        if not is_valid_splice(start_loc, delete_num, critical_indices):
            raise ValueError(
                f"Critical index was deleted during splice. This is not allowed: {text}, {start_loc}, {delete_num}"
            )
        index_modifier = (
            lambda i: i if i <= start_loc else i + len(insert_text) - delete_num
        )
        new_text = text[:start_loc] + insert_text + text[start_loc + delete_num :]
        return new_text, index_modifier

    if isinstance(task_sample, ArgumentsExtractionSample):
        critical_indices = [task_sample.task.trigger_loc]
        for frame_element in task_sample.frame_elements:
            critical_indices.append(frame_element.start_loc)
            critical_indices.append(frame_element.end_loc)
        new_text, index_modifier = modify_text(task_sample.task.text, critical_indices)
        return ArgumentsExtractionSample(
            frame_elements=[
                replace(
                    elm,
                    start_loc=index_modifier(elm.start_loc),
                    end_loc=index_modifier(elm.end_loc),
                )
                for elm in task_sample.frame_elements
            ],
            task=replace(
                task_sample.task,
                text=new_text,
                trigger_loc=index_modifier(task_sample.task.trigger_loc),
            ),
        )

    if isinstance(task_sample, FrameClassificationSample):
        critical_indices = [task_sample.task.trigger_loc]
        new_text, index_modifier = modify_text(task_sample.task.text, critical_indices)
        return FrameClassificationSample(
            frame=task_sample.frame,
            task=replace(
                task_sample.task,
                text=new_text,
                trigger_loc=index_modifier(task_sample.task.trigger_loc),
            ),
        )

    if isinstance(task_sample, TriggerIdentificationSample):
        critical_indices = task_sample.trigger_locs
        new_text, index_modifier = modify_text(task_sample.task.text, critical_indices)
        return TriggerIdentificationSample(
            trigger_locs=[index_modifier(loc) for loc in task_sample.trigger_locs],
            task=replace(task_sample.task, text=new_text),
        )

    raise ValueError(f"Unknown task sample type: {type(task_sample)}")



from __future__ import annotations

from frame_semantic_transformer.data.tasks import (
    ArgumentsExtractionSample,
    FrameClassificationSample,
    TaskSample,
    TriggerIdentificationSample,
)


def get_sample_text(sample: TaskSample) -> str:
    if isinstance(sample, ArgumentsExtractionSample):
        return sample.task.text
    if isinstance(sample, FrameClassificationSample):
        return sample.task.text
    if isinstance(sample, TriggerIdentificationSample):
        return sample.task.text
    raise ValueError(f"Unknown sample type: {type(sample)}")



from __future__ import annotations
from frame_semantic_transformer.data.augmentations.modification_helpers import (
    modify_text_without_changing_length,
)

from frame_semantic_transformer.data.tasks import TaskSample
from .DataAugmentation import DataAugmentation


class UppercaseAugmentation(DataAugmentation):
    def apply_augmentation(self, task_sample: TaskSample) -> TaskSample:
        def safe_uppercase(text: str) -> str:
            new_text = text.upper()
            # it turns out the some characters, like "ﬁ", become 2 chars when uppercased
            # just check to make sure we're not in that case here
            if len(new_text) != len(text):
                return text
            return new_text

        return modify_text_without_changing_length(task_sample, safe_uppercase)



from __future__ import annotations

import glob
import os


KEYS_TO_CHECK = {
    "val_loss": {"minimize": True},
    "val_args_extraction_f1": {"minimize": False},
    "val_trigger_identification_f1": {"minimize": False},
    "val_frame_classification_f1": {"minimize": False},
    "val_avg_f1": {"minimize": False},
}


def find_best_val_model_paths(outputs_dir: str) -> dict[str, str]:
    """
    Helper script to find the models with the higest validation scores by f1 or val loss after training
    """
    potential_outputs = [os.path.basename(obj) for obj in glob.glob(f"{outputs_dir}/*")]

    top_scores: dict[str, float] = {}
    best_models = {}
    for output_name in potential_outputs:
        if "epoch=" not in output_name:
            continue
        scores = get_model_scores(output_name)
        for key, value in scores.items():
            if (
                key not in top_scores
                or (KEYS_TO_CHECK[key]["minimize"] and value < top_scores[key])
                or (not KEYS_TO_CHECK[key]["minimize"] and value > top_scores[key])
            ):
                top_scores[key] = value
                best_models[key] = output_name
    return best_models


def get_model_scores(output_name: str) -> dict[str, float]:
    """
    Helper function to get the scores for a given model
    """
    scores = {}
    for name_part in output_name.split("-"):
        if "=" in name_part:
            key, value = name_part.split("=")
            if key in KEYS_TO_CHECK:
                scores[key] = float(value)
    if (
        "val_args_extraction_f1" in scores
        and "val_trigger_identification_f1" in scores
        and "val_frame_classification_f1" in scores
    ):
        scores["val_avg_f1"] = (
            scores["val_args_extraction_f1"]
            + scores["val_trigger_identification_f1"]
            + scores["val_frame_classification_f1"]
        ) / 3
    return scores



from __future__ import annotations

from dataclasses import dataclass


@dataclass
class Frame:
    """
    Representation of a FrameNet frame
    For training on your own data, you can use this class to represent your own frames
    """

    name: str
    core_elements: list[str]
    non_core_elements: list[str]
    lexical_units: list[str]


@dataclass
class FrameAnnotatedSentence:
    """
    Representation of a sentence with annotations for use in training
    If training on your own data, you'll need to create instances of this class for your training sentences
    """

    text: str
    annotations: list[FrameAnnotation]
    # if this text isn't annotated with every trigger loc, we shouldn't generate a trigger id task from it
    # but it's still useful for frame classification and argument extraction tasks
    skip_trigger_identification_task: bool = False


@dataclass
class FrameAnnotation:
    """
    A single frame occuring in a sentence
    """

    frame: str
    trigger_locs: list[int]
    frame_elements: list[FrameElementAnnotation]


@dataclass
class FrameElementAnnotation:
    """
    A single frame element in a frame annotation.
    Includes the name of the frame element and the start and end locations of the frame element in the sentence
    """

    name: str
    start_loc: int
    end_loc: int



from __future__ import annotations
import random
import string

from frame_semantic_transformer.data.augmentations.modification_helpers import (
    splice_text,
)
from frame_semantic_transformer.data.tasks import TaskSample
from .DataAugmentation import DataAugmentation, ProbabilityType
from .modification_helpers.get_sample_text import get_sample_text


class StripPunctuationAugmentation(DataAugmentation):

    max_to_remove: int
    min_to_remove: int

    def __init__(
        self,
        probability: ProbabilityType,
        max_to_remove: int = 5,
        min_to_remove: int = 1,
    ):
        self.max_to_remove = max_to_remove
        self.min_to_remove = min_to_remove
        super().__init__(probability)

    def apply_augmentation(self, task_sample: TaskSample) -> TaskSample:
        sample_text = get_sample_text(task_sample)
        punctuation_indices = find_punctuation_indices(sample_text)
        if len(punctuation_indices) == 0:
            return task_sample

        updated_sample = task_sample
        for _ in range(random.randint(self.min_to_remove, self.max_to_remove)):
            if len(punctuation_indices) == 0:
                break
            punctuation_index = random.choice(punctuation_indices)
            punctuation_indices.remove(punctuation_index)
            punctuation_indices = [
                i - 1 if i > punctuation_index else i for i in punctuation_indices
            ]
            try:
                updated_sample = splice_text(
                    updated_sample,
                    lambda _text, _critical_indices: (
                        punctuation_index,
                        1,
                        "",
                    ),
                )
            except ValueError:
                # The splice failed, so just return the sample
                return updated_sample
        return updated_sample


def find_punctuation_indices(text: str) -> list[int]:
    """
    Find the indices of all punctuation in the text.
    """
    # TODO: This would be more efficient with a regex
    return [i for i, char in enumerate(text) if char in string.punctuation]



from __future__ import annotations
from collections import defaultdict
import logging
from typing import Any, Type
from dataclasses import asdict, dataclass, field
from transformers import T5ForConditionalGeneration, T5TokenizerFast

from frame_semantic_transformer.constants import PADDING_LABEL_ID
from frame_semantic_transformer.data.LoaderDataCache import LoaderDataCache
from frame_semantic_transformer.data.data_utils import chunk_list
from frame_semantic_transformer.data.tasks.ArgumentsExtractionSample import (
    ArgumentsExtractionSample,
)
from frame_semantic_transformer.data.tasks.FrameClassificationSample import (
    FrameClassificationSample,
)
from frame_semantic_transformer.data.tasks.TaskSample import TaskSample
from frame_semantic_transformer.data.tasks.TriggerIdentificationSample import (
    TriggerIdentificationSample,
)
from frame_semantic_transformer.predict import predict_on_ids

logger = logging.getLogger(__name__)


def calc_eval_metrics(scores: EvalScores) -> dict[str, float]:
    """
    Calculate precision, recall, and f score
    Based on https://github.com/swabhs/open-sesame/blob/master/sesame/evaluation.py
    """
    if scores.true_pos == 0 and scores.false_pos == 0:
        precision = 0.0
    else:
        precision = scores.true_pos / (scores.true_pos + scores.false_pos)
    if scores.true_pos == 0 and scores.false_neg == 0:
        recall = 0.0
    else:
        recall = scores.true_pos / (scores.true_pos + scores.false_neg)
    if precision == 0 and recall == 0:
        f_score = 0.0
    else:
        f_score = 2.0 * precision * recall / (precision + recall)
    return {"precision": precision, "recall": recall, "f_score": f_score}


# TODO: figure out a better way to lookup the class from the string
TASK_SAMPLE_CLASS_MAP: dict[str, Type[TaskSample]] = {
    "args_extraction": ArgumentsExtractionSample,
    "frame_classification": FrameClassificationSample,
    "trigger_identification": TriggerIdentificationSample,
}


@dataclass
class EvalScores:
    true_pos: float = 0
    false_pos: float = 0
    false_neg: float = 0


@dataclass
class EvalFailure:
    input: str
    target: str
    predictions: list[str]


@dataclass
class TaskEvalResults:
    scores: EvalScores = field(default_factory=EvalScores)
    false_positives: list[EvalFailure] = field(default_factory=list)
    false_negatives: list[EvalFailure] = field(default_factory=list)

    def serialize(self) -> dict[str, Any]:
        return {
            "scores": asdict(self.scores),
            "false_positives": list(map(asdict, self.false_positives)),
            "false_negatives": list(map(asdict, self.false_negatives)),
        }


def evaluate_batch(
    model: T5ForConditionalGeneration,
    tokenizer: T5TokenizerFast,
    batch: Any,
    loader_cache: LoaderDataCache,
    predictions_per_sample: int = 5,
) -> dict[str, TaskEvalResults]:
    predictions = predict_on_ids(
        model,
        tokenizer,
        batch["input_ids"],
        batch["attention_mask"],
        skip_special_tokens=True,
        clean_up_tokenization_spaces=True,
        num_beams=predictions_per_sample,
        num_return_sequences=predictions_per_sample,
    )
    batched_predictions = chunk_list(predictions, predictions_per_sample)
    results: dict[str, TaskEvalResults] = defaultdict(TaskEvalResults)
    for preds, task, label, input_ids in zip(
        batched_predictions, batch["task"], batch["labels"], batch["input_ids"]
    ):
        assert len(preds) == predictions_per_sample
        target_tokens = [
            tok_id for tok_id in label.tolist() if tok_id != PADDING_LABEL_ID
        ]
        input_tokens = [
            tok_id for tok_id in input_ids.tolist() if tok_id != PADDING_LABEL_ID
        ]
        target = tokenizer.decode(
            target_tokens, skip_special_tokens=True, clean_up_tokenization_spaces=True
        )
        input = tokenizer.decode(
            input_tokens, skip_special_tokens=True, clean_up_tokenization_spaces=True
        )
        sample_class = TASK_SAMPLE_CLASS_MAP[task]
        true_pos, false_pos, false_neg = sample_class.evaluate_prediction(
            preds, target, input, loader_cache
        )
        results[task].scores.true_pos += true_pos
        results[task].scores.false_pos += false_pos
        results[task].scores.false_neg += false_neg
        if false_pos > 0:
            results[task].false_positives.append(
                EvalFailure(input=input, target=target, predictions=list(preds))
            )
        if false_neg > 0:
            results[task].false_negatives.append(
                EvalFailure(input=input, target=target, predictions=list(preds))
            )

    return results



from __future__ import annotations
import argparse
import logging
from transformers import T5ForConditionalGeneration, T5TokenizerFast

from frame_semantic_transformer.training import evaluate_model

logger = logging.getLogger(__name__)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Evaluate a semantic transformer model using FrameNet 1.7"
    )
    parser.add_argument(
        "--model-path",
        help="The path to the model to evaluate, or the name of a pretrained huggingface model",
    )
    parser.add_argument("--num-workers", type=int, default=2, help="default 2")
    parser.add_argument("--batch-size", type=int, default=8, help="default 8")
    args = parser.parse_args()

    logger.setLevel(logging.INFO)

    model = T5ForConditionalGeneration.from_pretrained(args.model_path)
    tokenizer = T5TokenizerFast.from_pretrained(args.model_path)

    evaluate_model(
        model,
        tokenizer,
        batch_size=args.batch_size,
        num_workers=args.num_workers,
    )



from __future__ import annotations
from frame_semantic_transformer.data.augmentations.modification_helpers import (
    modify_text_without_changing_length,
)

from frame_semantic_transformer.data.tasks import TaskSample
from .DataAugmentation import DataAugmentation


class LowercaseAugmentation(DataAugmentation):
    def apply_augmentation(self, task_sample: TaskSample) -> TaskSample:
        return modify_text_without_changing_length(task_sample, str.lower)



from __future__ import annotations
from abc import ABC, abstractmethod

from frame_semantic_transformer.data.augmentations.DataAugmentation import (
    DataAugmentation,
)
from frame_semantic_transformer.data.frame_types import Frame, FrameAnnotatedSentence


class InferenceLoader(ABC):
    """
    Base class for all inference loaders
    """

    def name(self) -> str:
        """
        Return a name for this loader.
        Frame-Semantic-Transformer will enforce that a loader with this name is used at inference time.
        """
        return self.__class__.__name__

    def strict_frame_elements(self) -> bool:
        """
        Return whether the loader strips out frame elements not in the frame definition.
        """
        return True

    def setup(self) -> None:
        """
        Perform any setup required, e.g. downloading needed data
        """
        pass

    @abstractmethod
    def load_frames(self) -> list[Frame]:
        """
        Load the full list of frames to be used during inference
        """
        pass

    @abstractmethod
    def normalize_lexical_unit_text(self, lu: str) -> str | set[str]:
        """
        Normalize a lexical unit like "takes.v" to "take".
        """
        pass

    def prioritize_lexical_unit(self, lu: str) -> bool:
        """
        Check if the lexical unit is relatively rare, so that it should be considered "high information"
        """
        return len(self.normalize_lexical_unit_text(lu)) >= 6


class TrainingLoader(ABC):
    """
    Base class for all training loaders
    """

    def name(self) -> str:
        """
        Return a name for this loader.
        Frame-Semantic-Transformer will enforce that a loader with this name is used at eval time.
        """
        return self.__class__.__name__

    def setup(self) -> None:
        """
        Perform any setup required, e.g. downloading needed data.
        """
        pass

    @abstractmethod
    def get_augmentations(self) -> list[DataAugmentation]:
        """
        Get a list of augmentations to apply to the training data
        """
        pass

    @abstractmethod
    def load_training_data(self) -> list[FrameAnnotatedSentence]:
        """
        Load the training data
        """
        pass

    @abstractmethod
    def load_validation_data(self) -> list[FrameAnnotatedSentence]:
        """
        Load the validation data
        """
        pass

    @abstractmethod
    def load_test_data(self) -> list[FrameAnnotatedSentence]:
        """
        Load the test data
        """
        pass



from __future__ import annotations

from typing import Any, Optional

import pytorch_lightning as pl
from torch.utils.data import DataLoader, Dataset

from frame_semantic_transformer.constants import DEFAULT_NUM_WORKERS


class TrainDataModule(pl.LightningDataModule):
    """
    Based on https://github.com/Shivanandroy/simpleT5/blob/main/simplet5/simplet5.py
    """

    batch_size: int
    train_dataset: Dataset[Any]
    val_dataset: Dataset[Any]
    test_dataset: Optional[Dataset[Any]]
    num_workers: int

    def __init__(
        self,
        train_dataset: Dataset[Any],
        val_dataset: Dataset[Any],
        test_dataset: Optional[Dataset[Any]] = None,
        batch_size: int = 8,
        num_workers: int = DEFAULT_NUM_WORKERS,
    ):
        super().__init__()
        self.batch_size = batch_size
        self.num_workers = num_workers

        self.train_dataset = train_dataset
        self.test_dataset = test_dataset
        self.val_dataset = val_dataset

    def train_dataloader(self) -> DataLoader[Any]:
        dataloader: Any = DataLoader(
            self.train_dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            shuffle=True,
        )
        return dataloader

    def val_dataloader(self) -> DataLoader[Any]:
        return DataLoader(
            self.val_dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            shuffle=False,
        )

    def test_dataloader(self) -> DataLoader[Any]:
        dataset = self.test_dataset if self.test_dataset else self.val_dataset
        return DataLoader(
            dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            shuffle=False,
        )



from __future__ import annotations
from typing import Any, Iterable, Optional

from nltk.corpus import framenet as fn

from frame_semantic_transformer.data.augmentations import (
    DoubleQuotesAugmentation,
    KeyboardAugmentation,
    LowercaseAugmentation,
    SimpleMisspellingAugmentation,
    RemoveEndPunctuationAugmentation,
    StripPunctuationAugmentation,
    SynonymAugmentation,
    UppercaseAugmentation,
)
from frame_semantic_transformer.data.augmentations.DataAugmentation import (
    DataAugmentation,
)
from frame_semantic_transformer.data.tasks import TriggerIdentificationSample
from .ensure_framenet_downloaded import ensure_framenet_downloaded
from .sesame_data_splits import SESAME_DEV_FILES, SESAME_TEST_FILES


from frame_semantic_transformer.data.frame_types import (
    FrameAnnotatedSentence,
    FrameAnnotation,
    FrameElementAnnotation,
)
from ..loader import TrainingLoader


def load_framenet_samples(
    include_docs: Optional[Iterable[str]] = None,
    exclude_docs: Optional[Iterable[str]] = None,
) -> list[FrameAnnotatedSentence]:
    samples: list[FrameAnnotatedSentence] = []
    for doc in fn.docs():
        if exclude_docs and doc["filename"] in exclude_docs:
            continue
        if include_docs and doc["filename"] not in include_docs:
            continue
        samples += parse_annotated_sentences_from_framenet_doc(doc)
    return samples


def load_framenet_samples_from_exemplars() -> list[FrameAnnotatedSentence]:
    samples: list[FrameAnnotatedSentence] = []
    # make sure we don't include exemplars if we've already included them in the training data
    all_doc_samples_text = {sample.text for sample in load_framenet_samples()}
    for sent in fn.exemplars():
        annotated_sent = parse_annotated_sentence_from_framenet_sentence(
            sent, skip_trigger_identification_task=True
        )
        if annotated_sent and annotated_sent.text not in all_doc_samples_text:
            samples.append(annotated_sent)
    return samples


def parse_annotated_sentences_from_framenet_doc(
    fn_doc: dict[str, Any]
) -> list[FrameAnnotatedSentence]:
    annotated_sentences = []
    for sentence in fn_doc["sentence"]:
        annotated_sentence = parse_annotated_sentence_from_framenet_sentence(sentence)
        if annotated_sentence:
            annotated_sentences.append(annotated_sentence)
    return annotated_sentences


def parse_annotated_sentence_from_framenet_sentence(
    fn_sentence: dict[str, Any],
    skip_trigger_identification_task: bool = False,
) -> FrameAnnotatedSentence | None:
    sentence_text = fn_sentence["text"]
    frame_annotations: list[FrameAnnotation] = []
    for fn_annotation in fn_sentence["annotationSet"]:
        if (
            "FE" in fn_annotation
            and "Target" in fn_annotation
            and "frame" in fn_annotation
        ):
            trigger_locs = [loc[0] for loc in fn_annotation["Target"]]
            # filter out broken annotations
            for trigger_loc in trigger_locs:
                if trigger_loc >= len(sentence_text):
                    return None
            frame_annotations.append(
                FrameAnnotation(
                    frame=fn_annotation["frame"]["name"],
                    trigger_locs=trigger_locs,
                    frame_elements=[
                        FrameElementAnnotation(
                            start_loc=fn_element[0],
                            end_loc=fn_element[1],
                            name=fn_element[2],
                        )
                        for fn_element in fn_annotation["FE"][0]
                    ],
                )
            )
    if len(frame_annotations) > 0:
        return FrameAnnotatedSentence(
            text=sentence_text,
            annotations=frame_annotations,
            skip_trigger_identification_task=skip_trigger_identification_task,
        )
    return None


class Framenet17TrainingLoader(TrainingLoader):
    include_exemplars: bool

    def __init__(self, include_exemplars: bool = False) -> None:
        super().__init__()
        self.include_exemplars = include_exemplars

    def setup(self) -> None:
        ensure_framenet_downloaded()

    def get_augmentations(self) -> list[DataAugmentation]:
        return [
            RemoveEndPunctuationAugmentation(0.5),
            DoubleQuotesAugmentation(0.2),
            StripPunctuationAugmentation(0.2),
            SynonymAugmentation(
                lambda sample: 0.2
                if isinstance(sample, TriggerIdentificationSample)
                else 0.05
            ),
            KeyboardAugmentation(
                lambda sample: 0.3
                if isinstance(sample, TriggerIdentificationSample)
                else 0.05
            ),
            SimpleMisspellingAugmentation(
                lambda sample: 0.3
                if isinstance(sample, TriggerIdentificationSample)
                else 0.05
            ),
            LowercaseAugmentation(0.1),
            UppercaseAugmentation(0.1),
        ]

    def load_training_data(self) -> list[FrameAnnotatedSentence]:
        training_samples = load_framenet_samples(
            exclude_docs=SESAME_DEV_FILES + SESAME_TEST_FILES
        )
        if self.include_exemplars:
            training_samples += load_framenet_samples_from_exemplars()
        return training_samples

    def load_test_data(self) -> list[FrameAnnotatedSentence]:
        return load_framenet_samples(include_docs=SESAME_TEST_FILES)

    def load_validation_data(self) -> list[FrameAnnotatedSentence]:
        return load_framenet_samples(include_docs=SESAME_DEV_FILES)



from __future__ import annotations
from dataclasses import dataclass
from typing import Sequence
from frame_semantic_transformer.data.LoaderDataCache import LoaderDataCache
from frame_semantic_transformer.data.data_utils import standardize_punct


from .Task import Task


@dataclass
class FrameClassificationTask(Task):
    text: str
    trigger_loc: int
    loader_cache: LoaderDataCache

    # -- input / target for training --

    @staticmethod
    def get_task_name() -> str:
        return "frame_classification"

    def get_input(self) -> str:
        potential_frames = self.loader_cache.get_possible_frames_for_trigger_bigrams(
            self.trigger_bigrams
        )
        return f"FRAME {' '.join(potential_frames)} : {self.trigger_labeled_text}"

    @staticmethod
    def parse_output(
        prediction_outputs: Sequence[str], loader_cache: LoaderDataCache
    ) -> str | None:
        for pred in prediction_outputs:
            if loader_cache.is_valid_frame(pred):
                # fix any capitalization differences between pred and the frame name
                return loader_cache.get_frame(pred).name
        return None

    # -- helper properties --

    @property
    def trigger_bigrams(self) -> list[list[str]]:
        """
        return bigrams of the trigger, trigger + next word, and prev word + trigger
        """
        pre_trigger_tokens = self.text[: self.trigger_loc].split()
        trigger_and_after_tokens = self.text[self.trigger_loc :].split()
        trigger = trigger_and_after_tokens[0]
        post_trigger_tokens = trigger_and_after_tokens[1:]
        bigrams: list[list[str]] = []
        if len(pre_trigger_tokens) > 0:
            bigrams.append([pre_trigger_tokens[-1], trigger])
        if len(post_trigger_tokens) > 0:
            bigrams.append([trigger, post_trigger_tokens[0]])
        # add the monogram last
        bigrams.append([trigger])
        return bigrams

    @property
    def trigger_labeled_text(self) -> str:
        pre_span = self.text[0 : self.trigger_loc]
        post_span = self.text[self.trigger_loc :]
        # TODO: handle these special chars better
        return standardize_punct(f"{pre_span}*{post_span}")



from __future__ import annotations
from typing import Optional

import torch
from torch.utils.data import DataLoader
from pytorch_lightning import Trainer
from transformers import T5ForConditionalGeneration, T5TokenizerFast
from frame_semantic_transformer.constants import DEFAULT_NUM_WORKERS

from frame_semantic_transformer.data.LoaderDataCache import LoaderDataCache
from frame_semantic_transformer.data.TaskSampleDataset import TaskSampleDataset
from frame_semantic_transformer.data.loaders.framenet17 import (
    Framenet17InferenceLoader,
    Framenet17TrainingLoader,
)
from frame_semantic_transformer.data.loaders.loader import (
    InferenceLoader,
    TrainingLoader,
)
from frame_semantic_transformer.data.tasks_from_annotated_sentences import (
    tasks_from_annotated_sentences,
)
from .TrainingModelWrapper import TrainingModelWrapper


def evaluate_model(
    model: T5ForConditionalGeneration,
    tokenizer: T5TokenizerFast,
    batch_size: int = 8,
    num_workers: int = DEFAULT_NUM_WORKERS,
    inference_loader: Optional[InferenceLoader] = None,
    training_loader: Optional[TrainingLoader] = None,
    log_eval_failures: bool = False,
) -> None:
    """
    Benchmark this model against the validation and test sets
    """

    if not inference_loader:
        inference_loader = Framenet17InferenceLoader()
    loader_cache = LoaderDataCache(inference_loader)
    if not training_loader:
        training_loader = Framenet17TrainingLoader()

    inference_loader.setup()
    training_loader.setup()

    config = model.config
    expected_inference_loader = (
        config.inference_loader
        if hasattr(config, "inference_loader")
        else inference_loader.name()
    )
    expected_training_loader = (
        config.training_loader
        if hasattr(config, "training_loader")
        else training_loader.name()
    )

    if expected_inference_loader != inference_loader.name():
        raise ValueError(
            f"Model was trained with inference loader {expected_inference_loader} but is being evaluated with {inference_loader.name()}"
        )
    if expected_training_loader != training_loader.name():
        raise ValueError(
            f"Model was trained with training loader {expected_training_loader} but is being evaluated with {training_loader.name()}"
        )

    model_wrapper = TrainingModelWrapper(
        model, tokenizer, loader_cache, log_eval_failures=log_eval_failures
    )
    trainer = Trainer(gpus=1, precision=32, max_epochs=1)

    val_dataset = TaskSampleDataset(
        tasks_from_annotated_sentences(
            training_loader.load_validation_data(), loader_cache
        ),
        tokenizer,
        balance_tasks=False,
    )
    test_dataset = TaskSampleDataset(
        tasks_from_annotated_sentences(training_loader.load_test_data(), loader_cache),
        tokenizer,
        balance_tasks=False,
    )

    with torch.no_grad():
        print("Evaluating on validation set")
        trainer.validate(
            model_wrapper,
            dataloaders=DataLoader(
                val_dataset, batch_size=batch_size, num_workers=num_workers
            ),
        )

        print("Evaluating on test set")
        trainer.test(
            model_wrapper,
            dataloaders=DataLoader(
                test_dataset, batch_size=batch_size, num_workers=num_workers
            ),
        )



from __future__ import annotations
from frame_semantic_transformer.data.LoaderDataCache import LoaderDataCache

from frame_semantic_transformer.data.loaders.loader import FrameAnnotatedSentence
from frame_semantic_transformer.data.tasks import (
    ArgumentsExtractionSample,
    ArgumentsExtractionTask,
    TaskSample,
    TriggerIdentificationSample,
    TriggerIdentificationTask,
    FrameClassificationSample,
    FrameClassificationTask,
)


def tasks_from_annotated_sentences(
    annotated_sentences: list[FrameAnnotatedSentence],
    loader_cache: LoaderDataCache,
) -> list[TaskSample]:
    task_samples: list[TaskSample] = []
    for annotated_sentence in annotated_sentences:
        trigger_locs = []
        for annotation in annotated_sentence.annotations:
            for trigger_loc in annotation.trigger_locs:
                trigger_locs.append(trigger_loc)
                task_samples.append(
                    FrameClassificationSample(
                        task=FrameClassificationTask(
                            text=annotated_sentence.text,
                            trigger_loc=trigger_loc,
                            loader_cache=loader_cache,
                        ),
                        frame=annotation.frame,
                    )
                )
                task_samples.append(
                    ArgumentsExtractionSample(
                        task=ArgumentsExtractionTask(
                            text=annotated_sentence.text,
                            trigger_loc=trigger_loc,
                            frame=annotation.frame,
                            loader_cache=loader_cache,
                        ),
                        frame_elements=annotation.frame_elements,
                    )
                )
        if not annotated_sentence.skip_trigger_identification_task:
            task_samples.append(
                TriggerIdentificationSample(
                    task=TriggerIdentificationTask(text=annotated_sentence.text),
                    trigger_locs=trigger_locs,
                )
            )
    return task_samples



from __future__ import annotations
from collections import defaultdict

from os import path
from glob import glob
import re

from nltk.corpus.reader.conll import ConllCorpusReader

from frame_semantic_transformer.data.augmentations import (
    KeyboardAugmentation,
    LowercaseAugmentation,
    RemoveEndPunctuationAugmentation,
    SimpleMisspellingAugmentation,
    SynonymAugmentation,
    UppercaseAugmentation,
)
from frame_semantic_transformer.data.augmentations.DataAugmentation import (
    DataAugmentation,
)

from frame_semantic_transformer.data.frame_types import (
    FrameAnnotatedSentence,
    FrameAnnotation,
    FrameElementAnnotation,
)
from ..loader import TrainingLoader
from .load_propbank_frames import load_propbank_frames


SPLITS = {
    "train": [
        "docs/evaluation/ewt.dev.txt",
        "docs/evaluation/ontonotes-train-list.txt",
    ],
    "val": ["docs/evaluation/ewt.dev.txt", "docs/evaluation/ontonotes-dev-list.txt"],
    "test": ["docs/evaluation/ewt.test.txt", "docs/evaluation/ontonotes-test-list.txt"],
}
EWT_GLOB = "data/google/ewt/**/*.gold_conll"
ONTONOTES_GLOB = "data/ontonotes/**/*.gold_conll"


def load_docs_set(base_path: str, docs_list_paths: list[str]) -> list[str]:
    docs_lookup = set()
    for docs_list_path in docs_list_paths:
        with open(path.join(base_path, docs_list_path)) as f:
            raw_docs = f.read().splitlines()
            # weirdly the ewt dev files end in .conllu but nothing else does
            docs_lookup.update([doc.replace(".conllu", "") for doc in raw_docs])

    docs = []
    ewt_docs = glob(path.join(base_path, EWT_GLOB), recursive=True)
    for doc in ewt_docs:
        doc_base = re.sub(r".*/data/google/ewt/", "", doc).replace(".gold_conll", "")
        if doc_base in docs_lookup:
            docs.append(doc)

    ontonotes_docs = glob(path.join(base_path, ONTONOTES_GLOB), recursive=True)
    for doc in ontonotes_docs:
        # for some reason the ontonotes list has 'ontonotes' in the path, but ewt doesn't have 'google/ewt'
        doc_base = re.sub(r".*/data/ontonotes/", "ontonotes/", doc).replace(
            ".gold_conll", ""
        )
        if doc_base in docs_lookup:
            docs.append(doc)
    return docs


def conll_word_index_to_locs(words: list[str], word_index: int) -> tuple[int, int]:
    """
    Take a list of words and an index of a word and return the start and end char indices of the word in the sentence
    """
    start_loc = 0
    for i, word in enumerate(words):
        if i == word_index:
            return start_loc, start_loc + len(word)
        start_loc += len(word) + 1
    raise ValueError("word index out of range")


def load_propbank_samples(
    docs_list: list[str], valid_frames: set[str]
) -> list[FrameAnnotatedSentence]:
    """
    Parse each of the propbank ontonotes and ewt gold conll files and return a list of FrameAnnotatedSentence objects
    """
    annotated_sentences = []
    for doc in docs_list:
        conll_reader = ConllCorpusReader(
            path.dirname(doc),
            path.basename(doc),
            ("ignore", "ignore", "ignore", "words", "pos", "tree", "srl"),
        )
        sents_map = defaultdict(list)
        for srl_instance in conll_reader.srl_instances():
            words = [word[0] for word in srl_instance.words]
            sentence = " ".join(words)
            frame_name = srl_instance.verb_stem
            if frame_name.lower() not in valid_frames:
                continue
            trigger_locs = [
                conll_word_index_to_locs(words, index)[0] for index in srl_instance.verb
            ]

            frame_elements = []
            for argument in srl_instance.arguments:
                words_range, frame_element_name = argument
                element_start_loc = conll_word_index_to_locs(words, words_range[0])[0]
                element_end_loc = conll_word_index_to_locs(words, words_range[1] - 1)[1]
                frame_elements.append(
                    FrameElementAnnotation(
                        frame_element_name, element_start_loc, element_end_loc
                    )
                )
            sents_map[sentence].append(
                FrameAnnotation(frame_name, trigger_locs, frame_elements)
            )

        for sentence, frame_annotations in sents_map.items():
            annotated_sentences.append(
                FrameAnnotatedSentence(sentence, frame_annotations)
            )
    return annotated_sentences


class Propbank34TrainingLoader(TrainingLoader):
    """
    This loader uses ontonotes and ewt data from propbank 3.1 to train a model
    You must clone https://github.com/propbank/propbank-release and set the propbank_release_dir to the path of the cloned repo
    You must also download the LDC data for ontonotes and ewt, and run map_all_to_conll.py as described in the propbank repo
    Sadly, this data isn't free so you'll need to get it yourself before working with this loader.
    """

    propbank_release_dir: str
    train_docs: list[str] = []
    val_docs: list[str] = []
    test_docs: list[str] = []
    valid_frames: set[str] = set()

    def __init__(self, propbank_release_dir: str) -> None:
        super().__init__()
        self.propbank_release_dir = propbank_release_dir

    def setup(self) -> None:
        self.valid_frames = {frame.name.lower() for frame in load_propbank_frames()}
        self.train_docs = load_docs_set(self.propbank_release_dir, SPLITS["train"])
        self.val_docs = load_docs_set(self.propbank_release_dir, SPLITS["val"])
        self.test_docs = load_docs_set(self.propbank_release_dir, SPLITS["test"])

    def get_augmentations(self) -> list[DataAugmentation]:
        return [
            RemoveEndPunctuationAugmentation(0.5),
            SynonymAugmentation(0.3),
            KeyboardAugmentation(0.3),
            SimpleMisspellingAugmentation(0.1),
            LowercaseAugmentation(0.1),
            UppercaseAugmentation(0.1),
        ]

    def load_training_data(self) -> list[FrameAnnotatedSentence]:
        return load_propbank_samples(self.train_docs, self.valid_frames)

    def load_test_data(self) -> list[FrameAnnotatedSentence]:
        return load_propbank_samples(self.test_docs, self.valid_frames)

    def load_validation_data(self) -> list[FrameAnnotatedSentence]:
        return load_propbank_samples(self.val_docs, self.valid_frames)



from __future__ import annotations
from collections import defaultdict
from functools import lru_cache
from itertools import product
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .frame_types import Frame
    from .loaders.loader import InferenceLoader


class LoaderDataCache:
    """
    Helper class which wraps a InferenceLoader and performs cached data lookups for performance
    """

    loader: InferenceLoader

    def __init__(self, loader: InferenceLoader):
        self.loader = loader

    def setup(self) -> None:
        """
        Perform any setup required, e.g. downloading needed data
        """
        self.loader.setup()

    @lru_cache(1)
    def get_frames_by_name(self) -> dict[str, Frame]:
        """
        cached fast-lookup helper for frame names -> frames for faster eval
        """
        results: dict[str, Frame] = {}
        for frame in self.loader.load_frames():
            results[normalize_name(frame.name)] = frame
        return results

    @lru_cache(1)
    def get_frame_element_name_loopkup(self) -> dict[str, str]:
        results: dict[str, str] = {}
        for frame in self.get_frames_by_name().values():
            for element in frame.core_elements:
                results[normalize_name(element)] = element
            for element in frame.non_core_elements:
                results[normalize_name(element)] = element
        return results

    def get_frame(self, name: str) -> Frame:
        """
        Get a frame by name
        """
        return self.get_frames_by_name()[normalize_name(name)]

    def is_valid_frame(self, name: str) -> bool:
        """
        Check if a frame name is valid
        """
        return normalize_name(name) in self.get_frames_by_name()

    def standardize_element_name(self, name: str) -> str | None:
        """
        Standardize a frame element name
        """
        norm_name = normalize_name(name)
        if norm_name not in self.get_frame_element_name_loopkup():
            return None if self.loader.strict_frame_elements() else name
        return self.get_frame_element_name_loopkup()[norm_name]

    @lru_cache(1)
    def get_lexical_unit_bigram_to_frame_lookup_map(self) -> dict[str, list[str]]:
        """
        Return a mapping of lexical unit bigrams to the list of frames they are associated with
        """
        uniq_lookup_map: dict[str, set[str]] = defaultdict(set)
        for frame in self.get_frames_by_name().values():
            for lu in frame.lexical_units:
                parts = lu.split()
                lu_bigrams: list[str] = []
                prev_part = None
                for part in parts:
                    # also key this as a mongram if there's only 1 element or the word is rare enough
                    if len(parts) == 1 or self.loader.prioritize_lexical_unit(part):
                        for norm_part in self._normalize_lexical_unit_ngram([part]):
                            lu_bigrams.append(norm_part)
                    if prev_part is not None:
                        for norm_parts in self._normalize_lexical_unit_ngram(
                            [prev_part, part]
                        ):
                            lu_bigrams.append(norm_parts)
                    prev_part = part

                for bigram in lu_bigrams:
                    uniq_lookup_map[bigram].add(frame.name)
        sorted_lookup_map: dict[str, list[str]] = {}
        for lu_bigram, frames in uniq_lookup_map.items():
            sorted_lookup_map[lu_bigram] = sorted(list(frames))
        return sorted_lookup_map

    def get_possible_frames_for_trigger_bigrams(
        self, bigrams: list[list[str]]
    ) -> list[str]:
        possible_frames = []
        lookup_map = self.get_lexical_unit_bigram_to_frame_lookup_map()
        for bigram in bigrams:
            # sorted here just to get a consistent ordering
            for normalized_bigram in sorted(self._normalize_lexical_unit_ngram(bigram)):
                if normalized_bigram in lookup_map:
                    bigram_frames = lookup_map[normalized_bigram]
                    possible_frames += bigram_frames
        # remove duplicates, while preserving order
        # https://stackoverflow.com/questions/1653970/does-python-have-an-ordered-set/53657523#53657523
        return list(dict.fromkeys(possible_frames))

    def _normalize_lexical_unit_ngram(self, ngram: list[str]) -> set[str]:
        norm_toks = [
            setify(self.loader.normalize_lexical_unit_text(tok)) for tok in ngram
        ]
        return {"_".join(combo) for combo in product(*norm_toks)}


def normalize_name(name: str) -> str:
    """
    Normalize a frame or element name to be lowercase and without underscores
    """
    return name.lower().replace("_", "")


def setify(input: str | set[str]) -> set[str]:
    """
    Convert a string or set to a set
    """
    if isinstance(input, str):
        return {input}
    return input



from .chain_augmentations import chain_augmentations
from .DataAugmentation import DataAugmentation
from .LowercaseAugmentation import LowercaseAugmentation
from .UppercaseAugmentation import UppercaseAugmentation
from .SimpleMisspellingAugmentation import SimpleMisspellingAugmentation
from .KeyboardAugmentation import KeyboardAugmentation
from .SynonymAugmentation import SynonymAugmentation
from .DoubleQuotesAugmentation import DoubleQuotesAugmentation
from .RemoveEndPunctuationAugmentation import RemoveEndPunctuationAugmentation
from .StripPunctuationAugmentation import StripPunctuationAugmentation

__all__ = (
    "chain_augmentations",
    "DataAugmentation",
    "DoubleQuotesAugmentation",
    "LowercaseAugmentation",
    "UppercaseAugmentation",
    "KeyboardAugmentation",
    "SimpleMisspellingAugmentation",
    "StripPunctuationAugmentation",
    "SynonymAugmentation",
    "RemoveEndPunctuationAugmentation",
)



from .modify_text_without_changing_length import modify_text_without_changing_length
from .splice_text import splice_text
from .get_sample_text import get_sample_text

__all__ = [
    "modify_text_without_changing_length",
    "splice_text",
    "get_sample_text",
]



from __future__ import annotations
from dataclasses import replace
from typing import Callable

from frame_semantic_transformer.data.tasks import (
    ArgumentsExtractionSample,
    FrameClassificationSample,
    TaskSample,
    TriggerIdentificationSample,
)


def modify_text_without_changing_length(
    task_sample: TaskSample, modify_text_cb: Callable[[str], str]
) -> TaskSample:
    """
    Helper to modify the text of a TaskSample without changing the length of the text
    This is a simple augmentation since it doesn't require rewriting indices

    This takes the task sample and a lambda function that takes the text of the task sample
    and returns the modified text. It then modifies the text of the task sample and returns
    """

    def modify_text(text: str) -> str:
        new_text = modify_text_cb(text)
        if len(new_text) != len(text):
            raise ValueError(
                f"Text length changed during augmentation: {text} -> {new_text}"
            )
        return new_text

    if isinstance(task_sample, ArgumentsExtractionSample):
        new_text = modify_text(task_sample.task.text)
        return replace(
            task_sample,
            task=replace(task_sample.task, text=new_text),
        )

    if isinstance(task_sample, FrameClassificationSample):
        new_text = modify_text(task_sample.task.text)
        return replace(
            task_sample,
            task=replace(task_sample.task, text=new_text),
        )

    if isinstance(task_sample, TriggerIdentificationSample):
        new_text = modify_text(task_sample.task.text)
        return replace(
            task_sample,
            task=replace(task_sample.task, text=new_text),
        )

    raise ValueError(f"Unknown task sample type: {type(task_sample)}")
